from absurdia import util
from absurdia.api_resources.abstract import ListableAPIResource
from absurdia import api_requestor

class Account(ListableAPIResource):
    OBJECT_NAME = "account"

    @classmethod
    def retrieve(cls, id=None, agent_token=None, **params):
        instance = cls(id, agent_token, **params)
        instance.refresh_from(instance.request("get", instance.class_url()))
        return instance

    @classmethod
    def current(cls, agent_token=None, absurdia_version=None, **params):
        requestor = api_requestor.APIRequestor(agent_token, api_version=absurdia_version)
        url = cls.class_url()
        response = requestor.request("get", url, params)
        return util.convert_to_absurdia_object(response, agent_token, absurdia_version)



import calendar
import datetime
import json
import platform
import time
from urllib.parse import urlencode, urlsplit, urlunsplit
import uuid
import warnings
from collections import OrderedDict

import absurdia
from absurdia import error, auth_error, http_client, version, util
from absurdia.absurdia_response import AbsurdiaResponse, AbsurdiaStreamResponse


def _encode_datetime(dttime):
    if dttime.tzinfo and dttime.tzinfo.utcoffset(dttime) is not None:
        utc_timestamp = calendar.timegm(dttime.utctimetuple())
    else:
        utc_timestamp = time.mktime(dttime.timetuple())

    return int(utc_timestamp)


def _encode_nested_dict(key, data, fmt="%s[%s]"):
    d = OrderedDict()
    for subkey, subvalue in data.items():
        d[fmt % (key, subkey)] = subvalue
    return d


def _api_encode(data):
    for key, value in data.items():
        key = util.utf8(key)
        if value is None:
            continue
        elif isinstance(value, list) or isinstance(value, tuple):
            for i, sv in enumerate(value):
                if isinstance(sv, dict):
                    subdict = _encode_nested_dict("%s[%d]" % (key, i), sv)
                    for k, v in _api_encode(subdict):
                        yield (k, v)
                else:
                    yield ("%s[%d]" % (key, i), util.utf8(sv))
        elif isinstance(value, dict):
            subdict = _encode_nested_dict(key, value)
            for subkey, subvalue in _api_encode(subdict):
                yield (subkey, subvalue)
        elif isinstance(value, datetime.datetime):
            yield (key, _encode_datetime(value))
        else:
            yield (key, util.utf8(value))


def _build_api_url(url, query):
    scheme, netloc, path, base_query, fragment = urlsplit(url)

    if base_query:
        query = "%s&%s" % (base_query, query)

    return urlunsplit((scheme, netloc, path, query, fragment))


class APIRequestor(object):
    def __init__(
        self,
        token=None,
        signature_key=None,
        client=None,
        api_base=None,
        api_version=None,
        account=None,
    ):
        self.api_base = api_base or absurdia.api_base
        self.agent_token = token
        self.agent_signature_key = signature_key
        self.api_version = api_version or absurdia.api_version
        self.absurdia_account = account

        if self.agent_token is None or self.agent_signature_key is None:
            util.load_agent()
        if self.agent_token is None:
            self.agent_token = absurdia.agent_token
        if self.agent_signature_key is None:
            self.agent_signature_key = absurdia.agent_signature_key

        self._default_proxy = None

        from absurdia import verify_ssl_certs as verify
        from absurdia import proxy

        if client:
            self._client = client
        elif absurdia.default_http_client:
            self._client = absurdia.default_http_client
            if proxy != self._default_proxy:
                warnings.warn(
                    "absurdia.proxy was updated after sending a "
                    "request - this is a no-op. To use a different proxy, "
                    "set absurdia.default_http_client to a new client "
                    "configured with the proxy."
                )
        else:
            # If the absurdia.default_http_client has not been set by the user
            # yet, we'll set it here. This way, we aren't creating a new
            # HttpClient for every request.
            absurdia.default_http_client = http_client.new_default_http_client(
                verify_ssl_certs=verify, proxy=proxy
            )
            self._client = absurdia.default_http_client
            self._default_proxy = proxy

    @classmethod
    def format_app_info(cls, info):
        str = info["name"]
        if info["version"]:
            str += "/%s" % (info["version"],)
        if info["url"]:
            str += " (%s)" % (info["url"],)
        return str

    def request(self, method, url, params=None, headers=None, body=None):
        rbody, rcode, rheaders, _ = self.request_raw(
            method.lower(), url, params, headers, body
        )
        resp = self.interpret_response(rbody, rcode, rheaders)
        return resp

    def handle_error_response(self, rbody, rcode, resp, rheaders):
        try:
            error_data = resp["error"]
        except (KeyError, TypeError):
            raise error.APIError(
                "Invalid response object from API: %r (HTTP response code "
                "was %d)" % (rbody, rcode),
                rbody,
                rcode,
                resp,
            )
            
        err = self.specific_api_error(
                rbody, rcode, resp, rheaders, error_data
            )

        raise err

    def specific_api_error(self, rbody, rcode, resp, rheaders, error_data):
        util.log_info(
            "Absurdia API error received",
            error_code=error_data.get("code"),
            error_type=error_data.get("error"),
            error_message=error_data,
            error_param=error_data.get("detail"),
        )

        # Rate limits were previously coded as 400's with code 'rate_limit'
        if rcode == 429:
            return error.RateLimitError(
                error_data.get("message"), rbody, rcode, resp, rheaders
            )
        elif rcode in [400, 404]:
            if error_data.get("type") == "idempotency_error":
                return error.IdempotencyError(
                    error_data.get("message"), rbody, rcode, resp, rheaders
                )
            else:
                return error.InvalidRequestError(
                    error_data.get("message"),
                    error_data.get("param"),
                    error_data.get("code"),
                    rbody,
                    rcode,
                    resp,
                    rheaders,
                )
        elif rcode == 401:
            return error.AuthenticationError(
                error_data.get("message"), rbody, rcode, resp, rheaders
            )
        elif rcode == 402:
            return error.CardError(
                error_data.get("message"),
                error_data.get("param"),
                error_data.get("code"),
                rbody,
                rcode,
                resp,
                rheaders,
            )
        elif rcode == 403:
            return error.PermissionError(
                error_data.get("message"), rbody, rcode, resp, rheaders
            )
        else:
            return error.APIError(
                error_data.get("message"), rbody, rcode, resp, rheaders
            )

    def specific_auth_error(self, rbody, rcode, resp, rheaders, error_code):
        description = resp.get("error_description", error_code)

        util.log_info(
            "Absurdia Auth error received",
            error_code=error_code,
            error_description=description,
        )

        args = [error_code, description, rbody, rcode, resp, rheaders]

        if error_code == "invalid_grant":
            return auth_error.InvalidGrantError(*args)
        elif error_code == "invalid_request":
            return auth_error.InvalidRequestError(*args)
        elif error_code == "unverified_account":
            return auth_error.UnverifiedAccount(*args)
        elif error_code == "signature_required":
            return auth_error.SignatureRequired(*args)
        elif error_code == "invalid_signature":
            return auth_error.InvalidSignature(*args)
        elif error_code == "system_error":
            return auth_error.SystemError(*args)
        return None

    def request_headers(self, agent_token, method):
        user_agent = "Absurdia/v1 PythonBindings/%s" % (version.VERSION,)
        if absurdia.app_info:
            user_agent += " " + self.format_app_info(absurdia.app_info)

        ua = {
            "bindings_version": version.VERSION,
            "lang": "python",
            "publisher": "absurdia",
            "httplib": self._client.name,
        }
        for attr, func in [
            ["lang_version", platform.python_version],
            ["platform", platform.platform],
            ["uname", lambda: " ".join(platform.uname())],
        ]:
            try:
                val = func()
            except Exception:
                val = "(disabled)"
            ua[attr] = val
        if absurdia.app_info:
            ua["application"] = absurdia.app_info

        headers = {
            "Abs-Client-User-Agent": json.dumps(ua),
            "User-Agent": user_agent,
            "Authorization": "Bearer %s" % (agent_token,),
            "Content-Type": "application/json",
            "Accept-Encoding": "br,gzip",
        }

        if method == "post":
            headers.setdefault("Idempotency-Key", str(uuid.uuid4()))

        if self.api_version is not None:
            headers["Absurdia-Version"] = self.api_version

        return headers

    def request_raw(
        self,
        method,
        url,
        params=None,
        supplied_headers=None,
        body=None,
    ):
        """
        Mechanism for issuing an API call
        """

        if self.agent_token:
            bearer = self.agent_token
        else:
            from absurdia import agent_token

            bearer = agent_token

        if bearer is None:
            raise error.AuthenticationError(
                "No agent key provided. (HINT: set your agent key using "
                '"absurdia.agent_token = <AGENT-KEY>"). You can generate agent keys '
                "from the Absurdia web interface.  See https://app.absurdia.eu/dash/agents "
                "for details, or email support@absurdia.eu if you have any "
                "questions."
            )

        abs_url = "%s%s" % (self.api_base, url)

        encoded_params = urlencode(list(_api_encode(params or {})))

        # Don't use strict form encoding by changing the square bracket control
        # characters back to their literals. This is fine by the server, and
        # makes these parameter strings easier to read.
        encoded_params = encoded_params.replace("%5B", "[").replace("%5D", "]")

        if method in ["get", "post", "patch", "delete"]:
            #post_data = encoded_params
            body = params
        else:
            raise error.APIConnectionError(
                "Unrecognized HTTP method %r.  This may indicate a bug in the "
                "Absurdia bindings.  Please contact support@absurdia.eu for "
                "assistance." % (method,)
            )

        headers = self.request_headers(bearer, method)
        if supplied_headers is not None:
            for key, value in supplied_headers:
                headers[key] = value

        util.log_info("Request to Absurdia api", method=method, path=abs_url)
        util.log_debug(
            "Post details",
            body=body,
            api_version=self.api_version,
        )
        rcontent, rcode, rheaders = self._client.request_with_retries(
                method, abs_url, headers, body=body, params=encoded_params
            )

        util.log_info("Absurdia API response", path=abs_url, response_code=rcode)
        util.log_debug("API response body", body=str(rcontent))

        return rcontent, rcode, rheaders, bearer

    def _should_handle_code_as_error(self, rcode):
        return not 200 <= rcode < 300

    def interpret_response(self, rbody, rcode, rheaders):
        try:
            if hasattr(rbody, "decode"):
                rbody = rbody.decode("utf-8")
            resp = AbsurdiaResponse(rbody, rcode, rheaders)
        except Exception:
            raise error.APIError(
                "Invalid response body from API: %s "
                "(HTTP response code was %d)" % (rbody, rcode),
                rbody,
                rcode,
                rheaders,
            )
        if self._should_handle_code_as_error(rcode):
            self.handle_error_response(rbody, rcode, resp.body, rheaders)
        return resp
 


import imp
import os

# Absurdia Python bindings

# Configuration variables

agent_filepath = "absurdia-agent.env"
agent_id = None
agent_token = None
agent_signature_key = None
fund_id = None
api_base = "https://api.absurdia.eu"
download_api_base = "https://data.absurdia.eu"
api_version = "v1"
verify_ssl_certs = True
proxy = None
default_http_client = None
app_info = None
enable_telemetry = True
max_network_retries = 0
ca_bundle_path = os.path.join(
    os.path.dirname(__file__), "data", "ca-certificates.crt"
)

# Set to either 'debug' or 'info', controls console logging
log = None

# API resources
from absurdia.api_resources import *  # noqa
from absurdia.markets import *  # noqa

# Takes a name and optional version and plugin URL.
def set_app_info(name, partner_id=None, url=None, version=None):
    global app_info
    app_info = {
        "name": name,
        "partner_id": partner_id,
        "url": url,
        "version": version,
    }


import json
from collections import OrderedDict

class AbsurdiaResponseBase(object):
    def __init__(self, code, headers):
        self.code = code
        self.headers = headers

    @property
    def idempotency_key(self):
        try:
            return self.headers["idempotency-key"]
        except KeyError:
            return None

    @property
    def request_id(self):
        try:
            return self.headers["x-request-id"]
        except KeyError:
            return None


class AbsurdiaResponse(AbsurdiaResponseBase):
    def __init__(self, body, code, headers):
        AbsurdiaResponseBase.__init__(self, code, headers)
        self.body = body
        self.data = json.loads(body)
        if code == 200 or code == 201:
            self.data = self.data['data']

class AbsurdiaStreamResponse(AbsurdiaResponseBase):
    def __init__(self, io, code, headers):
        AbsurdiaResponseBase.__init__(self, code, headers)
        self.io = io


from urllib.parse import quote_plus
from absurdia import api_requestor, error, util
from absurdia.absurdia_object import AbsurdiaObject

class APIResource(AbsurdiaObject):
    @classmethod
    def retrieve(cls, id=None, agent_token=None, **params):
        instance = cls(id, agent_token, **params)
        instance.refresh()
        return instance

    def refresh(self):
        id = self.get("id")
        if isinstance(id, str):
            self.refresh_from(self.request("get", self.class_url(), params={"id": id}))
        return self

    @classmethod
    def class_url(cls):
        if cls == APIResource:
            raise NotImplementedError(
                "APIResource is an abstract class.  You should perform "
                "actions on its subclasses (e.g. agent, fund)"
            )
        # Namespaces are separated in object names with periods (.) and in URLs
        # with forward slashes (/), so replace the former with the latter.
        base = cls.OBJECT_NAME.replace(".", "/")
        return "/v1/%ss" % (base,)

    # The `method_` and `url_` arguments are suffixed with an underscore to
    # avoid conflicting with actual request parameters in `params`.
    @classmethod
    def _static_request(
        cls,
        method_,
        url_,
        agent_token=None,
        idempotency_key=None,
        absurdia_version=None,
        absurdia_account=None,
        **params
    ):
        requestor = api_requestor.APIRequestor(
            agent_token, api_version=absurdia_version, account=absurdia_account
        )
        headers = util.populate_headers(idempotency_key)
        response = requestor.request(method_, url_, params, headers)
        return util.convert_to_absurdia_object(
            response, agent_token, absurdia_version, absurdia_account
        )

    # The `method_` and `url_` arguments are suffixed with an underscore to
    # avoid conflicting with actual request parameters in `params`.
    @classmethod
    def _static_request_stream(
        cls,
        method_,
        url_,
        agent_token=None,
        idempotency_key=None,
        absurdia_version=None,
        absurdia_account=None,
        **params
    ):
        requestor = api_requestor.APIRequestor(
            agent_token, api_version=absurdia_version, account=absurdia_account
        )
        headers = util.populate_headers(idempotency_key)
        response = requestor.request_stream(method_, url_, params, headers)
        return response


def custom_method(name, http_verb, http_path=None, is_streaming=False):
    if http_verb not in ["get", "post", "delete", "patch"]:
        raise ValueError(
            "Invalid http_verb: %s. Must be one of 'get', 'post', 'patch' or 'delete'"
            % http_verb
        )
    if http_path is None:
        http_path = name

    def wrapper(cls):
        def custom_method_request(cls, sid, **params):
            url = "%s/%s/%s" % (
                cls.class_url(),
                quote_plus(util.utf8(sid)),
                http_path,
            )
            obj = cls._static_request(http_verb, url, **params)

            # For list objects, we have to attach the parameters so that they
            # can be referenced in auto-pagination and ensure consistency.
            if "object" in obj and obj.object == "list":
                obj._retrieve_params = params

            return obj

        def custom_method_request_stream(cls, sid, **params):
            url = "%s/%s/%s" % (
                cls.class_url(),
                quote_plus(util.utf8(sid)),
                http_path,
            )
            return cls._static_request_stream(http_verb, url, **params)

        if is_streaming:
            class_method_impl = classmethod(custom_method_request_stream)
        else:
            class_method_impl = classmethod(custom_method_request)

        existing_method = getattr(cls, name, None)
        if existing_method is None:
            setattr(cls, name, class_method_impl)
        else:
            # If a method with the same name we want to use already exists on
            # the class, we assume it's an instance method. In this case, the
            # new class method is prefixed with `_cls_`, and the original
            # instance method is decorated with `util.class_method_variant` so
            # that the new class method is called when the original method is
            # called as a class method.
            setattr(cls, "_cls_" + name, class_method_impl)
            instance_method = util.class_method_variant("_cls_" + name)(
                existing_method
            )
            setattr(cls, name, instance_method)

        return cls

    return wrapper


class CreateableAPIResource(APIResource):
    @classmethod
    def create(
        cls,
        agent_token=None,
        idempotency_key=None,
        absurdia_version=None,
        absurdia_account=None,
        **params
    ):
        requestor = api_requestor.APIRequestor(
            agent_token, api_version=absurdia_version, account=absurdia_account
        )
        url = cls.class_url()
        headers = util.populate_headers(idempotency_key)
        response = requestor.request("post", url, params, headers)

        return util.convert_to_absurdia_object(
            response, agent_token, absurdia_version, absurdia_account
        )


class ListableAPIResource(APIResource):
    @classmethod
    def auto_paging_iter(cls, *args, **params):
        return cls.list(*args, **params).auto_paging_iter()

    @classmethod
    def list(
        cls, agent_token=None, absurdia_account=None, **params
    ):
        requestor = api_requestor.APIRequestor(agent_token, account=absurdia_account,)
        url = cls.class_url()
        response = requestor.request("get", url, params)
        absurdia_object = util.convert_to_absurdia_object(response, agent_token, absurdia_account=absurdia_account)
        return absurdia_object


import sys
import textwrap
import ssl
import time
import random
import threading
import json

import absurdia
import absurdia.util as util
import absurdia.error as error

try:
    import httpx
    version = httpx.__version__
    major, minor, patch = [int(i) for i in version.split(".")]
    if (major, minor, patch) < (0, 21, 2):
        sys.stderr.write(
            "Warning: the Absurdia library requires that your Python "
            '"httpx" library be newer than version 0.21.2, but your '
            '"httpx" library is version %s. Absurdia will fall back to '
            'the alternate HTTP library "requests" so everything should work. We '
            'recommend upgrading your "httpx" library. If you have any '
            "questions, please contact support@absurdia.eu. (HINT: running "
            '"pip install -U httpx" should upgrade your httpx '
            "library to the latest version.)" % (version,)
        )
        httpx = None
except ImportError:
    httpx = None
    try:
        import requests
    except ImportError:
        requests = None

def _now_ms():
    return int(round(time.time() * 1000))

def new_default_http_client(*args, **kwargs):
    if httpx:
        impl = HttpxClient
    elif requests:
        impl = RequestsClient
    else:
        sys.stderr.write(
            "Error: the Absurdia library requires that the Python "
            '"httpx" or "requests" library be installed in your environment, but none '
            'was found. Run "pip install httpx" tp install the preferred HTTP client. '
            "If you have any questions, please contact support@absurdia.eu."
        )
        raise ImportError("Missing dependency: httpx or requests")

    return impl(*args, **kwargs)

class RequestMetrics(object):
    def __init__(self, request_id, request_duration_ms):
        self.request_id = request_id
        self.request_duration_ms = request_duration_ms

    def payload(self):
        return {
            "request_id": self.request_id,
            "request_duration_ms": self.request_duration_ms,
        }


class HTTPClient(object):
    MAX_DELAY = 2
    INITIAL_DELAY = 0.5
    MAX_RETRY_AFTER = 60

    def __init__(self, verify_ssl_certs=True, proxy=None):
        self._verify_ssl_certs = verify_ssl_certs
        if proxy:
            if isinstance(proxy, str):
                proxy = {"http": proxy, "https": proxy}
            if not isinstance(proxy, dict):
                raise ValueError(
                    "Proxy(ies) must be specified as either a string "
                    "URL or a dict() with string URL under the"
                    " "
                    "https"
                    " and/or "
                    "http"
                    " keys."
                )
        self._proxy = proxy.copy() if proxy else None

        self._thread_local = threading.local()

    def request_with_retries(self, method, url, headers, params=None, body=None):
        return self._request_with_retries_internal(
            method, url, params, headers, body, is_streaming=False
        )

    def request_stream_with_retries(
        self, method, url, headers, post_data=None
    ):
        return self._request_with_retries_internal(
            method, url, headers, post_data, is_streaming=True
        )

    def _request_with_retries_internal(
        self, method, url, params, headers, body, is_streaming
    ):
        self._add_telemetry_header(headers)

        num_retries = 0

        while True:
            request_start = _now_ms()

            try:
                if is_streaming:
                    response = self.request_stream(
                        method, url, params, headers, body
                    )
                else:
                    response = self.request(method, url, params, headers, body)
                connection_error = None
            except error.APIConnectionError as e:
                connection_error = e
                response = None

            if self._should_retry(response, connection_error, num_retries):
                if connection_error:
                    util.log_info(
                        "Encountered a retryable error %s"
                        % connection_error.user_message
                    )
                num_retries += 1
                sleep_time = self._sleep_time_seconds(num_retries, response)
                util.log_info(
                    (
                        "Initiating retry %i for request %s %s after "
                        "sleeping %.2f seconds."
                        % (num_retries, method, url, sleep_time)
                    )
                )
                time.sleep(sleep_time)
            else:
                if response is not None:
                    self._record_request_metrics(response, request_start)

                    return response
                else:
                    raise connection_error

    def request(self, method, url, headers, post_data=None):
        raise NotImplementedError(
            "HTTPClient subclasses must implement `request`"
        )

    def request_stream(self, method, url, headers, post_data=None):
        raise NotImplementedError(
            "HTTPClient subclasses must implement `request_stream`"
        )

    def _should_retry(self, response, api_connection_error, num_retries):
        if num_retries >= self._max_network_retries():
            return False

        if response is None:
            # We generally want to retry on timeout and connection
            # exceptions, but defer this decision to underlying subclass
            # implementations. They should evaluate the driver-specific
            # errors worthy of retries, and set flag on the error returned.
            return api_connection_error.should_retry

        _, status_code, rheaders = response

        # The API may ask us not to retry (eg; if doing so would be a no-op)
        # or advise us to retry (eg; in cases of lock timeouts); we defer to that.
        #
        # Note that we expect the headers object to be a CaseInsensitiveDict, as is the case with the requests library.
        if rheaders is not None and "abs-should-retry" in rheaders:
            if rheaders["abs-should-retry"] == "false":
                return False
            if rheaders["abs-should-retry"] == "true":
                return True

        # Retry on conflict errors.
        if status_code == 409:
            return True

        # Retry on 500, 503, and other internal errors.
        #
        # Note that we expect the abs-should-retry header to be false
        # in most cases when a 500 is returned, since our idempotency framework
        # would typically replay it anyway.
        if status_code >= 500:
            return True

        return False

    def _max_network_retries(self):
        from absurdia import max_network_retries

        # Configured retries, isolated here for tests
        return max_network_retries

    def _retry_after_header(self, response=None):
        if response is None:
            return None
        _, _, rheaders = response

        try:
            return int(rheaders["retry-after"])
        except (KeyError, ValueError):
            return None

    def _sleep_time_seconds(self, num_retries, response=None):
        # Apply exponential backoff with initial_network_retry_delay on the
        # number of num_retries so far as inputs.
        # Do not allow the number to exceed max_network_retry_delay.
        sleep_seconds = min(
            HTTPClient.INITIAL_DELAY * (2 ** (num_retries - 1)),
            HTTPClient.MAX_DELAY,
        )

        sleep_seconds = self._add_jitter_time(sleep_seconds)

        # But never sleep less than the base sleep seconds.
        sleep_seconds = max(HTTPClient.INITIAL_DELAY, sleep_seconds)

        # And never sleep less than the time the API asks us to wait, assuming it's a reasonable ask.
        retry_after = self._retry_after_header(response) or 0
        if retry_after <= HTTPClient.MAX_RETRY_AFTER:
            sleep_seconds = max(retry_after, sleep_seconds)

        return sleep_seconds

    def _add_jitter_time(self, sleep_seconds):
        # Randomize the value in [(sleep_seconds/ 2) to (sleep_seconds)]
        # Also separated method here to isolate randomness for tests
        sleep_seconds *= 0.5 * (1 + random.uniform(0, 1))
        return sleep_seconds

    def _add_telemetry_header(self, headers):
        last_request_metrics = getattr(
            self._thread_local, "last_request_metrics", None
        )
        if absurdia.enable_telemetry and last_request_metrics:
            telemetry = {
                "last_request_metrics": last_request_metrics.payload()
            }
            headers["Abs-Client-Telemetry"] = json.dumps(telemetry)

    def _record_request_metrics(self, response, request_start):
        _, _, rheaders = response
        if "Abs-Request-Id" in rheaders and absurdia.enable_telemetry:
            request_id = rheaders["Abs-Request-Id"]
            request_duration_ms = _now_ms() - request_start
            self._thread_local.last_request_metrics = RequestMetrics(
                request_id, request_duration_ms
            )

    def close(self):
        raise NotImplementedError(
            "HTTPClient subclasses must implement `close`"
        )

class HttpxClient(HTTPClient):
    name = "httpx"

    def __init__(self, timeout=80, session=None, **kwargs):
        super(HttpxClient, self).__init__(**kwargs)
        self._session = session
        self._timeout = timeout

    def request(self, method, url, headers, params, body=None):
        return self._request_internal(
            method, url, params, headers, body, is_streaming=False
        )

    def _request_internal(self, method, url, headers, params, body, is_streaming):
        kwargs = {}
        if self._proxy:
            kwargs["proxies"] = self._proxy

        if is_streaming:
            kwargs["stream"] = True

        if getattr(self._thread_local, "session", None) is None:
            self._thread_local.session = self._session or httpx.Client(http2=True)

        try:
            try:
                result = self._thread_local.session.request(
                    method,
                    url,
                    headers=headers,
                    params=params,
                    json=body,
                    timeout=self._timeout,
                    **kwargs
                )
            except TypeError as e:
                raise TypeError(
                    "Warning: It looks like your installed version of the "
                    '"httpx" library is not compatible with Absurdia\'s '
                    "usage thereof. (HINT: The most likely cause is that "
                    'your "httpx" library is out of date. You can fix '
                    'that by running "pip install -U httpx".) The '
                    "underlying error was: %s" % (e,)
                )

            if is_streaming:
                content = result.raw
            else:
                # This causes the content to actually be read, which could cause
                # e.g. a socket timeout. TODO: The other fetch methods probably
                # are susceptible to the same and should be updated.
                content = result.content

            status_code = result.status_code
        except Exception as e:
            # Would catch just requests.exceptions.RequestException, but can
            # also raise ValueError, RuntimeError, etc.
            self._handle_request_error(e)
        return content, status_code, result.headers

    def _handle_request_error(self, e):

        # Catch SSL error first as it belongs to ConnectionError,
        # but we don't want to retry
        if isinstance(e, ssl.SSLCertVerificationError):
            msg = (
                "Could not verify Absurdia's SSL certificate.  Please make "
                "sure that your network is not intercepting certificates.  "
                "If this problem persists, let us know at "
                "support@absurdia.eu."
            )
            err = "%s: %s" % (type(e).__name__, str(e))
            should_retry = False
        # Retry only timeout and connect errors; similar to urllib3 Retry
        elif isinstance(
            e,
            (httpx.Client, httpx.ConnectError),
        ):
            msg = (
                "Unexpected error communicating with Absurdia.  "
                "If this problem persists, let us know at "
                "support@absurdia.eu."
            )
            err = "%s: %s" % (type(e).__name__, str(e))
            should_retry = True
        # Catch remaining request exceptions
        elif isinstance(e, httpx.ConnectError):
            msg = (
                "Unexpected error communicating with Absurdia.  "
                "If this problem persists, let us know at "
                "support@absurdia.eu."
            )
            err = "%s: %s" % (type(e).__name__, str(e))
            should_retry = False
        else:
            msg = (
                "Unexpected error communicating with Absurdia. "
                "It looks like there's probably a configuration "
                "issue locally.  If this problem persists, let us "
                "know at support@absurdia.eu."
            )
            err = "A %s was raised" % (type(e).__name__,)
            if str(e):
                err += " with error message %s" % (str(e),)
            else:
                err += " with no error message"
            should_retry = False

        msg = textwrap.fill(msg) + "\n\n(Network error: %s)" % (err,)
        raise error.APIConnectionError(msg, should_retry=should_retry)

    def close(self):
        if getattr(self._thread_local, "session", None) is not None:
            self._thread_local.session.close()

class RequestsClient(HTTPClient):
    name = "requests"

    def __init__(self, timeout=80, session=None, **kwargs):
        super(RequestsClient, self).__init__(**kwargs)
        self._session = session
        self._timeout = timeout

    def request(self, method, url, headers, params=None, body=None):
        return self._request_internal(
            method, url, params, headers, body, is_streaming=False
        )

    def _request_internal(self, method, url, params, headers, body, is_streaming):
        kwargs = {}
        if self._verify_ssl_certs:
            kwargs["verify"] = absurdia.ca_bundle_path
        else:
            kwargs["verify"] = False

        if self._proxy:
            kwargs["proxies"] = self._proxy

        if is_streaming:
            kwargs["stream"] = True

        if getattr(self._thread_local, "session", None) is None:
            self._thread_local.session = self._session or requests.Session()

        try:
            try:
                util.log_debug("Request: %s %s %s" % (method, url, headers))
                result = self._thread_local.session.request(
                    method,
                    url,
                    headers=headers,
                    params=params,
                    json=body,
                    timeout=self._timeout,
                    **kwargs
                )
            except TypeError as e:
                raise TypeError(
                    "Warning: It looks like your installed version of the "
                    '"requests" library is not compatible with Absurdia\'s '
                    "usage thereof. (HINT: The most likely cause is that "
                    'your "requests" library is out of date. You can fix '
                    'that by running "pip install -U requests".) The '
                    "underlying error was: %s" % (e,)
                )

            if is_streaming:
                content = result.raw
            else:
                # This causes the content to actually be read, which could cause
                # e.g. a socket timeout. TODO: The other fetch methods probably
                # are susceptible to the same and should be updated.
                content = result.content

            status_code = result.status_code
        except Exception as e:
            # Would catch just requests.exceptions.RequestException, but can
            # also raise ValueError, RuntimeError, etc.
            self._handle_request_error(e)
        return content, status_code, result.headers

    def _handle_request_error(self, e):

        # Catch SSL error first as it belongs to ConnectionError,
        # but we don't want to retry
        if isinstance(e, requests.exceptions.SSLError):
            msg = (
                "Could not verify Absurdia's SSL certificate.  Please make "
                "sure that your network is not intercepting certificates.  "
                "If this problem persists, let us know at "
                "support@absurdia.eu."
            )
            err = "%s: %s" % (type(e).__name__, str(e))
            should_retry = False
        # Retry only timeout and connect errors; similar to urllib3 Retry
        elif isinstance(
            e,
            (requests.exceptions.Timeout, requests.exceptions.ConnectionError),
        ):
            msg = (
                "Unexpected error communicating with Absurdia.  "
                "If this problem persists, let us know at "
                "support@absurdia.eu."
            )
            err = "%s: %s" % (type(e).__name__, str(e))
            should_retry = True
        # Catch remaining request exceptions
        elif isinstance(e, requests.exceptions.RequestException):
            msg = (
                "Unexpected error communicating with Absurdia.  "
                "If this problem persists, let us know at "
                "support@absurdia.eu."
            )
            err = "%s: %s" % (type(e).__name__, str(e))
            should_retry = False
        else:
            msg = (
                "Unexpected error communicating with Absurdia. "
                "It looks like there's probably a configuration "
                "issue locally.  If this problem persists, let us "
                "know at support@absurdia.eu."
            )
            err = "A %s was raised" % (type(e).__name__,)
            if str(e):
                err += " with error message %s" % (str(e),)
            else:
                err += " with no error message"
            should_retry = False

        msg = textwrap.fill(msg) + "\n\n(Network error: %s)" % (err,)
        raise error.APIConnectionError(msg, should_retry=should_retry)

    def close(self):
        if getattr(self._thread_local, "session", None) is not None:
            self._thread_local.session.close()

 


from absurdia.markets.abstract import MarketResource

class Trade(MarketResource):
    OBJECT_NAME = "trade"
    LATEST_TRADES = {}
    
    def __init__(self, symbol: str):
        super().__init__(symbol)
        Trade.LATEST_TRADES[symbol] = self

    def get(self):
        return self.refresh()

def get_latest_trades(symbol: str):
    return Trade(symbol=symbol)


from absurdia.absurdia_object import AbsurdiaObject
from absurdia import api_requestor, util
from absurdia.absurdia_response import AbsurdiaResponse

class MarketResource():
    def __init__(self, requestor=None, cache=None, **params):
        if requestor:
            self.__requestor = requestor
        else:
            self.__requestor = api_requestor.APIRequestor(**params)
        self.__retrieved_at = 0
        self.value = None
        self.id = None
        self.symbol = None
        if cache is None:
            self.__cache = 0
        else:
            self.__cache = cache

    @classmethod
    def class_url(cls):
        if cls == MarketResource:
            raise NotImplementedError(
                "MarketResource is an abstract class.  You should perform "
                "actions on its subclasses (e.g. agent, fund)"
            )
        # Namespaces are separated in object names with periods (.) and in URLs
        # with forward slashes (/), so replace the former with the latter.
        base = cls.OBJECT_NAME.replace(".", "/")
        return "/v1/%ss" % (base,)
    
    def get_(self, symbol=None, force=False, **params):
        if self.__retrieved_at < (util.current_timestamp('ms') - self.__cache) \
        or force:
            url = self.class_url()
            if symbol:
                symbol = {"symbol": symbol}
            self.value = self.__requestor.request("get", url, params=symbol, headers=None, body=params)
        return self.value.data



import functools
import io
import logging
import sys
import os
import re
import datetime
from urllib.parse import quote_plus

try:
    pd = __import__("pandas")
except ImportError:
    pd = None

import absurdia

ABSURDIA_LOG = os.environ.get("ABSURDIA_LOG", default="warn")

logger = logging.getLogger("absurdia")

__all__ = [
    "io",
    "utf8",
    "log_info",
    "log_debug",
    "dashboard_link",
    "logfmt",
]

SUPPORTED_ASSETS = [
    "BTC",
    "ETH",
    "BUSD",
    "USDT",
    "USDC",
    "USD",
    "EUR",
    "GBP",
    "CHF"
]

SUPPORTED_MARKET_TYPES = ["SPOT"]

SUPPORTED_VENUES = ["BIN", "FTX", "BPD"]

def current_timestamp(granularity="us"):
    if granularity == "us":
        return int(datetime.datetime.now().timestamp() * 1e6)
    elif granularity == "ms":
        return int(datetime.datetime.now().timestamp() * 1e3)
    elif granularity == "s":
        return int(datetime.datetime.now().timestamp())
    else:
        raise ValueError("Invalid granularity: {}".format(granularity))

def to_df(obj):
    if pd is None:
        raise ImportError("`pandas` is required to convert to a DataFrame. Install with `pip install pandas`")
    else:
        return pd.DataFrame(obj, columns=obj[0].keys())

def utf8(value):
    return value.encode("utf-8")

def is_appengine_dev():
    return "APPENGINE_RUNTIME" in os.environ and "Dev" in os.environ.get(
        "SERVER_SOFTWARE", ""
    )

def get_object_classes():
    # This is here to avoid a circular dependency
    from absurdia.object_classes import OBJECT_CLASSES

    return OBJECT_CLASSES

def _console_log_level():
    if absurdia.log in ["debug", "info"]:
        return absurdia.log
    elif ABSURDIA_LOG in ["debug", "info"]:
        return ABSURDIA_LOG
    else:
        return None


def log_debug(message, **params):
    msg = logfmt(dict(message=message, **params))
    if _console_log_level() == "debug":
        print(msg, file=sys.stderr)
    logger.debug(msg)


def log_info(message, **params):
    msg = logfmt(dict(message=message, **params))
    if _console_log_level() in ["debug", "info"]:
        print(msg, file=sys.stderr)
    logger.info(msg)

def logfmt(props):
    def fmt(key, val):
        return u"{key}={val}".format(key=str(key), val=str(val))

    return u" ".join([fmt(key, val) for key, val in sorted(props.items())])

# Load agent credentials
def load_agent():
    if os.environ.get('ABSURDIA_TOKEN') is not None:
        absurdia.agent_token = os.environ['ABSURDIA_TOKEN']
    if os.environ.get('ABSURDIA_SIG_KEY') is not None:
        absurdia.agent_signature_key = os.environ['ABSURDIA_SIG_KEY']
    if os.environ.get('ABSURDIA_AGENT_ID') is not None:
        absurdia.agent_id = os.environ['ABSURDIA_AGENT_ID']
    if os.path.exists(absurdia.agent_filepath) \
    and (absurdia.agent_token is None \
    or absurdia.agent_signature_key is None \
    or absurdia.agent_id is None):
        file = open(absurdia.agent_filepath, "r").read()
        idx = file.find("ABSURDIA_TOKEN")
        if idx > -1 and absurdia.agent_token is None:
            idx = idx + (len("ABSURDIA_TOKEN") + 1)
            absurdia.agent_token = file[idx:file.find("\n", idx)]
        idx = file.find("ABSURDIA_SIG_KEY")
        if idx > -1 and absurdia.agent_signature_key is None:
            idx = idx + (len("ABSURDIA_SIG_KEY") + 1)
            absurdia.agent_signature_key = file[idx:file.find("\n", idx)]
        idx = file.find("ABSURDIA_AGENT_ID")
        if idx > -1 and absurdia.agent_id is None:
            idx = idx + (len("ABSURDIA_AGENT_ID") + 1)
            absurdia.agent_id = file[idx:file.find("\n", idx)]

def convert_to_absurdia_object(
    resp, agent_token=None, absurdia_version=None, absurdia_account=None, klass_name=None
):
    # If we get a AbsurdiaResponse, we'll want to return a
    # AbsurdiaObject with the last_response field filled out with
    # the raw API response information
    absurdia_response = None

    if isinstance(resp, absurdia.absurdia_response.AbsurdiaResponse):
        absurdia_response = resp
        resp = absurdia_response.data

    if isinstance(resp, list):
        return list([
            convert_to_absurdia_object(
                i, agent_token, absurdia_version, absurdia_account, klass_name
            )
            for i in resp
        ])
    elif isinstance(resp, dict) and not isinstance(
        resp, absurdia.absurdia_object.AbsurdiaObject
    ):
        resp = resp.copy()
        klass_name = resp.get("object") if klass_name is None else klass_name 
        if isinstance(klass_name, str):
            klass = get_object_classes().get(
                klass_name, absurdia.absurdia_object.AbsurdiaObject
            )
        else:
            klass = absurdia.absurdia_object.AbsurdiaObject

        return klass.construct_from(
            resp,
            agent_token,
            absurdia_version=absurdia_version,
            absurdia_account=absurdia_account,
            last_response=absurdia_response,
        )
    else:
        return resp

def convert_to_dict(obj):
    """Converts a AbsurdiaObject back to a regular dict.
    Nested AbsurdiaObject are also converted back to regular dicts.
    :param obj: The AbsurdiaObject to convert.
    :returns: The AbsurdiaObject as a dict.
    """
    if isinstance(obj, list):
        return [convert_to_dict(i) for i in obj]
    # This works by virtue of the fact that AbsurdiaObject _are_ dicts. The dict
    # comprehension returns a regular dict and recursively applies the
    # conversion to each value.
    elif isinstance(obj, dict):
        return {k: convert_to_dict(v) for k, v in iter(obj)}
    else:
        return obj


def populate_headers(idempotency_key):
    if idempotency_key is not None:
        return {"Idempotency-Key": idempotency_key}
    return None


def merge_dicts(x, y):
    z = x.copy()
    z.update(y)
    return z


def sanitize_id(id):
    utf8id = utf8(id)
    quotedId = quote_plus(utf8id)
    return quotedId


class class_method_variant(object):
    def __init__(self, class_method_name):
        self.class_method_name = class_method_name

    def __call__(self, method):
        self.method = method
        return self

    def __get__(self, obj, objtype=None):
        @functools.wraps(self.method)
        def _wrapper(*args, **kwargs):
            if obj is not None:
                # Method was called as an instance method, e.g.
                # instance.method(...)
                return self.method(obj, *args, **kwargs)
            elif len(args) > 0 and isinstance(args[0], objtype):
                # Method was called as a class method with the instance as the
                # first argument, e.g. Class.method(instance, ...) which in
                # Python is the same thing as calling an instance method
                return self.method(args[0], *args[1:], **kwargs)
            else:
                # Method was called as a class method, e.g. Class.method(...)
                class_method = getattr(objtype, self.class_method_name)
                return class_method(*args, **kwargs)

        return _wrapper

def validate_symbol(symbol: str) -> bool:
    symbol = symbol.upper()
    parts = symbol.split("-")
    if len(parts) < 2:
        return False
    elif len(parts) == 2:
        venue = None
    elif len(parts) == 3:
        venue = parts[2]
    assets = parts[0].split(".")
    market_type = parts[1]

    if len(assets) < 2:
        return False
    base = assets[0]
    quote = assets[1]
    if base == quote:
        return False

    if not base in SUPPORTED_ASSETS or not quote in SUPPORTED_ASSETS:
        return False
    
    if not market_type in SUPPORTED_MARKET_TYPES:
        return False

    if not venue in SUPPORTED_VENUES:
        return False

    return True

def compose_symbol(base: str, quote: str, market_type: str, venue: str) -> str:
    base = base.upper()
    quote = quote.upper()
    market_type = market_type.upper()
    venue = venue.upper()
    if not base in SUPPORTED_ASSETS or not quote in SUPPORTED_ASSETS:
        return ""
    if not market_type in SUPPORTED_MARKET_TYPES:
        return ""
    if not venue in SUPPORTED_VENUES:
        return ""
    return "{base}.{quote}-{market_type}-{venue}".format(base, quote, market_type, venue)


from absurdia import util
from absurdia import api_requestor
from absurdia.markets.abstract import MarketResource

class Orderbook(MarketResource):
    OBJECT_NAME = "orderbook"
    RETRIEVED_AT = 0
    LATEST_ORDERBOOK = None
    def __init__(self, symbol: str):
        super().__init__(symbol)

    @classmethod
    def current(
        cls, api_key=None, absurdia_version=None, absurdia_account=None, force_refresh=False, **params
    ):
        if Orderbook.RETRIEVED_AT < (util.current_timestamp('ms') - 100) \
        or force_refresh:
            Orderbook.RETRIEVED_AT = util.current_timestamp('s')
            requestor = api_requestor.APIRequestor(
                api_key, api_version=absurdia_version, account=absurdia_account
            )
            url = cls.class_url()
            response, api_key = requestor.request("get", url, params)
            response = util.convert_to_absurdia_object(
                response, api_key, absurdia_version, absurdia_account
            )
            Orderbook.LATEST_ORDERBOOK = response
        return Orderbook.LATEST_ORDERBOOK
    
    @classmethod
    def get(cls, symbol=None, **params):
        if cls.LATEST_ORDERBOOK is None:
            cls.current(symbol=symbol, **params)

def get_latest_orderbook(symbol: str):
    return Orderbook(symbol=symbol).current()


from absurdia.markets.abstract import MarketResource

class Price(MarketResource):
    OBJECT_NAME = "price"
    LATEST_PRICES = {}
    
    def __init__(self, symbol: str):
        super().__init__(symbol)
        Price.LATEST_PRICES[symbol] = self

    def get(self):
        return self.refresh()

def get_latest_price(symbol: str):
    #price = Price.LATEST_PRICES.get(symbol)
    #if price:
    #    return price.refresh()
    #else:
    #    return Price(symbol).refresh()
    return Price.retrieve(symbol=symbol)


from absurdia.markets.symbol import Symbol
from absurdia.markets.price import Price
from absurdia.markets.orderbook import Orderbook

# Convenience functions
from absurdia.markets.symbol import get_symbols as symbols
from absurdia.markets.symbol import get_symbol as symbol
from absurdia.markets.price import get_latest_price as price
from absurdia.markets.trade import get_latest_trades as trades
from absurdia.markets.orderbook import get_latest_orderbook as orderbook




import absurdia
from absurdia.error import AbsurdiaError


class AuthError(AbsurdiaError):
    def __init__(
        self,
        code,
        description,
        http_body=None,
        http_status=None,
        json_body=None,
        headers=None,
    ):
        super(AuthError, self).__init__(
            description, http_body, http_status, json_body, headers, code
        )

    def construct_error_object(self):
        if self.json_body is None:
            return None

        return (
            absurdia.api_resources.error_object.AuthErrorObject.construct_from(
                self.json_body, absurdia.agent_token
            )
        )

class InvalidGrantError(AuthError):
    pass


class InvalidRequestError(AuthError):
    pass


class UnverifiedAccount(AuthError):
    pass


class SignatureRequired(AuthError):
    pass


class InvalidSignature(AuthError):
    pass

class SystemError(AuthError):
    pass


from absurdia import util
from absurdia.markets.abstract import MarketResource

class Symbol(MarketResource):
    OBJECT_NAME = "symbol"
    LATEST_SYMBOLS = {}

    def __init__(self, symbol: str = None, with_response=None, **params):
        super().__init__(cache=3600000, **params)
        if symbol is None:
            for symbol in self.get_():
                primary = symbol.get("symbol")
                Symbol(symbol = primary, with_response=symbol)
        else:
            symbol = symbol.upper()
            if not util.validate_symbol(symbol):
                raise ValueError("Invalid symbol")

            if symbol in Symbol.LATEST_SYMBOLS:
                self.__value = Symbol.LATEST_SYMBOLS[symbol]
                self.__primary = symbol
            else:
                if with_response is None:
                    self.__value = self.get_(symbol)[0]
                else:
                    self.__value = with_response
                self.__primary = symbol
                Symbol.LATEST_SYMBOLS[symbol] = self.__value

    def __str__(self):
        return self.__primary

    def __repr__(self):
        return "Symbol('{primary}')".format(primary=self.__primary)

    @property
    def base(self):
        return self.__value.get("base")
    
    @property
    def quote(self):
        return self.__value.get("quote")

    @property
    def market_type(self):
        return self.__value.get("market_type")
    
    @property
    def venue(self):
        return self.__value.get("exchange")

    @property
    def available_services(self):
        return self.__value.get("services")

    def refresh(self):
        super().get_(self.__primary)[0]

    @classmethod
    def get_all(cls):
        return list(cls.LATEST_SYMBOLS.values())
        
def get_symbols(**params):
    Symbol(**params)
    return Symbol.get_all()

def get_symbol(symbol: str = None, base=None, quote=None, venue=None, **params):
    if symbol is not None:
        return Symbol(symbol=symbol, **params)
    elif base is not None and quote is not None and venue is not None:
        return Symbol(base=base, quote=quote, venue=venue, **params)
